SequenceableCollection subclass: #SparseArray
	instanceVariableNames: 'ranges size defaultValue'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Spreadsheet-Logic'!

!SparseArray methodsFor: 'accessing' stamp: 'hmm 9/24/2007 17:12'!
at: index
	ranges do: [:each | (each includesIndex: index) ifTrue: [^ each at: index]].
	^defaultValue! !

!SparseArray methodsFor: 'accessing' stamp: 'hmm 9/29/2007 13:18'!
at: index put: value
	| i |
	ranges do: [:each |
		(each includesIndex: index) ifTrue: [^ each at: index put: value]].
	ranges do: [:each |
		(each hasAdjacentIndex: index) ifTrue: [
			each at: index put: value.
			self coalesceRanges.
			^value]].
	i := ranges findFirst: [:each | each firstIndex > index].
	i = 0 ifTrue: [i := ranges size+1].
	ranges := (Array new: ranges size+1)
		replaceFrom: 1 to: i-1 with: ranges startingAt: 1;
		at: i put: (SparseArrayRange new index: index value: value);
		replaceFrom: i+1 to: ranges size+1 with: ranges startingAt: i;
		yourself.
	^value! !

!SparseArray methodsFor: 'accessing' stamp: 'hmm 9/24/2007 20:38'!
defaultValue
	^defaultValue! !

!SparseArray methodsFor: 'accessing' stamp: 'hmm 9/24/2007 20:38'!
defaultValue: anObject
	defaultValue := anObject! !

!SparseArray methodsFor: 'accessing' stamp: 'hmm 9/24/2007 20:32'!
size
	^size! !


!SparseArray methodsFor: 'initialize-release' stamp: 'hmm 9/24/2007 17:13'!
initSize: anInteger
	size := anInteger.
	ranges := Array new! !


!SparseArray methodsFor: 'private' stamp: 'hmm 9/24/2007 17:14'!
coalesceRanges
	"This is called when two ranges might be adjacent and could be coalesced. Only one such pair can exist at any time"
	| first next |
	first := ranges first.
	2 to: ranges size do: [:i |
		next := ranges at: i.
		(first coalesce: next) ifTrue: [
			ranges := ranges copyWithout: next.
			^self]]! !


!SparseArray methodsFor: 'copying' stamp: 'hmm 9/24/2007 21:09'!
postCopy
	ranges := ranges collect: [:each | each copy]! !

!SparseArray methodsFor: 'copying' stamp: 'hmm 9/24/2007 21:10'!
shallowCopy
	^self clone! !


!SparseArray methodsFor: 'enumerating' stamp: 'hmm 9/29/2007 13:20'!
nonEmptyElementsDo: aBlock
	ranges do: [:each |
		each values do: aBlock]! !

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

SparseArray class
	instanceVariableNames: ''!

!SparseArray class methodsFor: 'instance creation' stamp: 'hmm 9/24/2007 17:22'!
new: size
	^self basicNew initSize: size! !

!SparseArray class methodsFor: 'instance creation' stamp: 'hmm 9/24/2007 20:38'!
new: size withAll: anObject
	^(self new: size) defaultValue: anObject; yourself! !


Object subclass: #SparseArrayRange
	instanceVariableNames: 'firstIndex lastIndex values'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Spreadsheet-Logic'!

!SparseArrayRange methodsFor: 'testing' stamp: 'hmm 9/24/2007 17:04'!
hasAdjacentIndex: index
	^index + 1 = firstIndex or: [index - 1 = lastIndex]! !

!SparseArrayRange methodsFor: 'testing' stamp: 'hmm 9/24/2007 17:03'!
hasIndex: anInteger
	^anInteger between: firstIndex and: lastIndex! !

!SparseArrayRange methodsFor: 'testing' stamp: 'hmm 9/24/2007 17:12'!
includesIndex: anInteger
	^anInteger between: firstIndex and: lastIndex! !


!SparseArrayRange methodsFor: 'accessing' stamp: 'hmm 9/24/2007 17:04'!
at: index
	^values at: index-firstIndex+1! !

!SparseArrayRange methodsFor: 'accessing' stamp: 'hmm 9/24/2007 17:11'!
at: index put: anObject
	index+1 = firstIndex ifTrue: [
		firstIndex := index.
		values := values copyWithFirst: anObject.
		^anObject].
	index-1 = lastIndex ifTrue: [
		lastIndex := index.
		values := values copyWith: anObject.
		^anObject].
	^values at: index-firstIndex+1 put: anObject! !

!SparseArrayRange methodsFor: 'accessing' stamp: 'hmm 9/24/2007 17:04'!
firstIndex
	^firstIndex! !

!SparseArrayRange methodsFor: 'accessing' stamp: 'hmm 9/24/2007 17:04'!
lastIndex
	^lastIndex! !

!SparseArrayRange methodsFor: 'accessing' stamp: 'hmm 9/24/2007 20:37'!
values
	^values! !


!SparseArrayRange methodsFor: 'coalescing' stamp: 'hmm 9/24/2007 17:05'!
coalesce: aRange
	"Merge the two ranges if there is no remaining space between them. Return true if merge happened."
	lastIndex +1 = aRange firstIndex ifFalse: [^false].
	values := values, aRange values.
	lastIndex := aRange lastIndex.
	^true! !


!SparseArrayRange methodsFor: 'initialize-release' stamp: 'hmm 9/24/2007 17:07'!
index: index value: anObject
	firstIndex := lastIndex := index.
	values := Array with: anObject! !


!SparseArrayRange methodsFor: 'printing' stamp: 'hmm 9/24/2007 20:35'!
printOn: aStream
	super printOn: aStream.
	aStream space; print: firstIndex; nextPutAll: '..'; print: lastIndex! !


!SparseArrayRange methodsFor: 'copying' stamp: 'hmm 9/24/2007 21:13'!
postCopy
	values := values copy! !


Model subclass: #Spreadsheet
	instanceVariableNames: 'rows columns cachedRowIndex cachedRowTop cachedColumnIndex cachedColumnLeft waves'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Spreadsheet-Logic'!

!Spreadsheet methodsFor: 'initialize-release' stamp: 'hmm 10/11/2007 08:17'!
initialize
	columns := SparseArray new: 256 withAll: (SpreadsheetColumn new sheet: self; width: 70).
	rows := SparseArray new: 16384 withAll: (SpreadsheetRow new sheet: self; height: 15).
	waves := Dictionary new! !


!Spreadsheet methodsFor: 'accessing' stamp: 'hmm 9/30/2007 21:20'!
cellAt: aPoint
	^(rows at: aPoint y) at: aPoint x! !

!Spreadsheet methodsFor: 'accessing' stamp: 'hmm 11/3/2007 11:33'!
cellRuleOrValueStringAt: aPoint
	| cell |
	cell := (self cellAt: aPoint).
	^cell rule isNil
		ifTrue: [cell valueString]
		ifFalse: ['=', cell ruleString]! !

!Spreadsheet methodsFor: 'accessing' stamp: 'hmm 11/2/2007 23:10'!
cellStringAt: aPoint
	^(self cellAt: aPoint) valueString! !

!Spreadsheet methodsFor: 'accessing' stamp: 'hmm 9/30/2007 22:15'!
cellStringAt: aPoint put: aString
	(self realCellAt: aPoint) valueString: aString.
	self calculate! !

!Spreadsheet methodsFor: 'accessing' stamp: 'hmm 9/24/2007 20:53'!
columns
	^columns! !

!Spreadsheet methodsFor: 'accessing' stamp: 'hmm 10/2/2007 19:33'!
realCellAt: aPoint
	| cell r c |
	cell := (r := rows at: aPoint y) at: aPoint x.
	cell isDummy ifTrue: [
		cell := cell copy point: aPoint.
		r isDummy ifTrue: [
			rows at: aPoint y put: (r := r copy)].
		r at: aPoint x put: cell.
		c := columns at: aPoint x.
		c isDummy ifTrue: [
			columns at: aPoint x put: (c copy index: aPoint x)]].
	^cell! !

!Spreadsheet methodsFor: 'accessing' stamp: 'hmm 9/24/2007 21:17'!
rowAt: index
	| row |
	row := rows at: index.
	row isDummy ifTrue: [row := row copy index: index].
	^row! !

!Spreadsheet methodsFor: 'accessing' stamp: 'hmm 9/24/2007 20:53'!
rows
	^rows! !

!Spreadsheet methodsFor: 'accessing' stamp: 'hmm 9/24/2007 20:56'!
valueAt: aPoint
	^(self cellAt: aPoint) value! !

!Spreadsheet methodsFor: 'accessing' stamp: 'hmm 9/24/2007 20:57'!
valueAt: aPoint put: anObject
	| cell |
	cell := self realCellAt: aPoint.
	cell value: anObject! !


!Spreadsheet methodsFor: 'accessing-geometry' stamp: 'hmm 9/29/2007 09:23'!
columnLeftAt: index
	index = cachedColumnIndex ifTrue: [^cachedColumnLeft].
	(cachedColumnIndex = nil or: [cachedColumnIndex > (index*2)]) ifTrue: [
		"search from beginning"
		cachedColumnLeft := 0.
		1 to: index-1 do: [:i | cachedColumnLeft := cachedColumnLeft + (columns at: i) width].
		cachedColumnIndex := index.
		^cachedColumnLeft].
	[cachedColumnIndex > index] whileTrue: [
		cachedColumnIndex := cachedColumnIndex-1.
		cachedColumnLeft := cachedColumnLeft - (columns at: cachedColumnIndex) width].
	[cachedColumnIndex < index] whileTrue: [
		cachedColumnLeft := cachedColumnLeft + (columns at: cachedColumnIndex) width.
		cachedColumnIndex := cachedColumnIndex+1].
	^cachedColumnLeft! !

!Spreadsheet methodsFor: 'accessing-geometry' stamp: 'hmm 9/25/2007 13:31'!
columnRightAt: index
	^self columnLeftAt: index+1! !

!Spreadsheet methodsFor: 'accessing-geometry' stamp: 'hmm 10/3/2007 19:48'!
cornerForCellPoint: aPoint
	^(self columnRightAt: aPoint x) @ (self rowBottomAt: aPoint y)! !

!Spreadsheet methodsFor: 'accessing-geometry' stamp: 'hmm 9/30/2007 21:39'!
originForCellPoint: aPoint
	^(self columnLeftAt: aPoint x) @ (self rowTopAt: aPoint y)! !

!Spreadsheet methodsFor: 'accessing-geometry' stamp: 'hmm 9/25/2007 13:30'!
rowBottomAt: index
	^self rowTopAt: index+1! !

!Spreadsheet methodsFor: 'accessing-geometry' stamp: 'hmm 9/25/2007 13:29'!
rowTopAt: index
	index = cachedRowIndex ifTrue: [^cachedRowTop].
	(cachedRowIndex = nil or: [cachedRowIndex > (index*2)]) ifTrue: [
		"search from beginning"
		cachedRowTop := 0.
		1 to: index-1 do: [:i | cachedRowTop := cachedRowTop + (rows at: i) height].
		cachedRowIndex := index.
		^cachedRowTop].
	[cachedRowIndex > index] whileTrue: [
		cachedRowIndex := cachedRowIndex-1.
		cachedRowTop := cachedRowTop - (rows at: cachedRowIndex) height].
	[cachedRowIndex < index] whileTrue: [
		cachedRowTop := cachedRowTop + (rows at: cachedRowIndex) height.
		cachedRowIndex := cachedRowIndex+1].
	^cachedRowTop! !

!Spreadsheet methodsFor: 'accessing-geometry' stamp: 'hmm 9/27/2007 14:33'!
visibleColumnsIn: aRectangle
	| x first last |
	x := 0.
	first := (1 to: columns size)
			detect: [:i |
				x := x + (columns at: i) width.
				x > aRectangle left]
			ifNone: [0].
	first = 0 ifTrue: [^1 to: 0].
	x > aRectangle right ifTrue: [^first to: first].
	last := (first+1 to: columns size)
			detect: [:i |
				x := x + (columns at: i) width.
				x > aRectangle right]
			ifNone: [columns size].
	^first to: last! !

!Spreadsheet methodsFor: 'accessing-geometry' stamp: 'hmm 9/25/2007 13:43'!
visibleRowsIn: aRectangle
	| y first last |
	y := 0.
	first := (1 to: rows size)
			detect: [:i |
				y := y + (rows at: i) height.
				y > aRectangle top]
			ifNone: [0].
	first = 0 ifTrue: [^1 to: 0].
	y > aRectangle bottom ifTrue: [^first to: first].
	last := (first+1 to: rows size)
			detect: [:i |
				y := y + (rows at: i) height.
				y > aRectangle bottom]
			ifNone: [rows size].
	^first to: last! !



!Spreadsheet methodsFor: 'calculating' stamp: 'hmm 10/3/2007 20:57'!
calculate
	| lastW w |
	lastW := -1.
	[waves isEmpty] whileFalse: [
		w := waves keys asSortedCollection first.
		w <= lastW ifTrue: [self error: 'dependency circle'].
		lastW := w.
		(waves removeKey: w) do: [:cell | cell recomputeValue]]! !

!Spreadsheet methodsFor: 'calculating' stamp: 'hmm 9/29/2007 15:03'!
defineCalculationOrder
	| todo nextTodo |
	todo := OrderedCollection new.
	self cellsDo: [:cell |
		cell hasPrerequisites
			ifTrue: [
				cell detachFromPrerequisites.
				todo add: cell]
			ifFalse: [cell wave: 0]].
	[todo isEmpty] whileFalse: [
		nextTodo := OrderedCollection new.
		todo do: [:cell |
			cell computeWave ifFalse: [nextTodo add: cell]].
		nextTodo size = todo size ifTrue: ["break circular dependeny" self halt].
		todo := nextTodo]! !

!Spreadsheet methodsFor: 'calculating' stamp: 'hmm 10/3/2007 20:56'!
markForRecalculation: aCell
	waves isNil ifTrue: [waves := Dictionary new].
	(waves at: aCell wave ifAbsentPut: [Set new]) add: aCell! !

!Spreadsheet methodsFor: 'calculating' stamp: 'hmm 9/30/2007 21:33'!
recalculate
	waves := Dictionary new.
	self cellsDo: [:each |
		each rule isNil ifFalse: [self markForRecalculation: each]].
	self calculate! !


!Spreadsheet methodsFor: 'enumerating' stamp: 'hmm 9/30/2007 21:12'!
cellsDo: aBlock
	rows nonEmptyElementsDo: [:row | row cells nonEmptyElementsDo: aBlock]! !


!Spreadsheet methodsFor: 'changing' stamp: 'hmm 11/10/2007 12:33'!
changedCell: aCell
	self changed: #display with: (aCell point corner: aCell point)! !

!Spreadsheet methodsFor: 'changing' stamp: 'hmm 10/3/2007 23:29'!
changedColumn: aColumn
	cachedColumnIndex := cachedColumnLeft := nil.
	self changed: #displayFull! !

!Spreadsheet methodsFor: 'changing' stamp: 'hmm 11/10/2007 12:34'!
changedRow: aColumn
	cachedRowIndex := cachedRowTop := nil.
	self changed: #displayFull! !


Model subclass: #SpreadsheetCell
	instanceVariableNames: 'sheet value rule rowIndex columnIndex wave'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Spreadsheet-Logic'!

!SpreadsheetCell methodsFor: 'initialize-release' stamp: 'hmm 9/24/2007 21:42'!
sheet: aSpreadsheet
	sheet := aSpreadsheet.
	rowIndex := columnIndex := 0! !


!SpreadsheetCell methodsFor: 'printing' stamp: 'hmm 10/2/2007 19:39'!
printOn: aStream
	aStream nextPut: $[.
	columnIndex printAsColumnRefOn: aStream.
	rowIndex printOn: aStream.
	aStream space.
	value printOn: aStream.
	rule isNil ifFalse: [
		aStream space; nextPut: $=.
		rule printOn: aStream for: self precedence: 0].
	aStream nextPut: $]! !


!SpreadsheetCell methodsFor: 'accessing' stamp: 'hmm 9/24/2007 21:02'!
column
	^sheet columns at: columnIndex! !

!SpreadsheetCell methodsFor: 'accessing' stamp: 'hmm 9/29/2007 14:58'!
columnIndex
	^columnIndex! !

!SpreadsheetCell methodsFor: 'accessing' stamp: 'hmm 9/30/2007 21:23'!
computeWave
	| max |
	max := 0.
	self prerequisiteCells do: [:each |
		each wave isNil ifTrue: [^false].
		max := max max: each wave].
	self prerequisiteCells do: [:each | each addDependent: self].
	wave := max + 1.
	^true! !

!SpreadsheetCell methodsFor: 'accessing' stamp: 'hmm 9/24/2007 21:01'!
point
	^columnIndex @ rowIndex! !

!SpreadsheetCell methodsFor: 'accessing' stamp: 'hmm 9/24/2007 21:01'!
point: aPoint
	columnIndex := aPoint x.
	rowIndex := aPoint y! !

!SpreadsheetCell methodsFor: 'accessing' stamp: 'hmm 9/30/2007 21:16'!
prerequisiteCells
	rule isNil ifTrue: [^#()].
	^rule cellReferences collect: [:each | each referenceInCell: self]! !

!SpreadsheetCell methodsFor: 'accessing' stamp: 'hmm 9/24/2007 21:02'!
row
	^sheet rows at: rowIndex! !

!SpreadsheetCell methodsFor: 'accessing' stamp: 'hmm 9/29/2007 14:57'!
rowIndex
	^rowIndex! !

!SpreadsheetCell methodsFor: 'accessing' stamp: 'hmm 9/29/2007 13:28'!
rule
	^rule! !

!SpreadsheetCell methodsFor: 'accessing' stamp: 'hmm 9/30/2007 22:30'!
rule: aRuleOrNil
	rule = aRuleOrNil ifTrue: [^self].
	self detachFromPrerequisites.
	rule := aRuleOrNil.
	sheet defineCalculationOrder.
	rule = nil ifFalse: [sheet markForRecalculation: self]! !

!SpreadsheetCell methodsFor: 'accessing' stamp: 'hmm 11/3/2007 11:32'!
ruleString
	| str |
	^rule isNil ifTrue: [''] ifFalse: [
		str := WriteStream on: String new.
		rule printOn: str for: self precedence: 0.
		str contents]! !

!SpreadsheetCell methodsFor: 'accessing' stamp: 'hmm 9/30/2007 22:14'!
ruleString: aString
	self rule: (SpreadsheetRuleParser new parse: aString forCell: self)! !

!SpreadsheetCell methodsFor: 'accessing' stamp: 'hmm 9/30/2007 21:20'!
sheet
	^sheet! !

!SpreadsheetCell methodsFor: 'accessing' stamp: 'hmm 9/24/2007 21:02'!
value
	^value! !

!SpreadsheetCell methodsFor: 'accessing' stamp: 'hmm 11/10/2007 12:33'!
value: anObject
	(value class = anObject class and: [value = anObject])
		ifFalse: [
			value := anObject.
			self changed: #value.
			sheet changedCell: self]! !

!SpreadsheetCell methodsFor: 'accessing' stamp: 'hmm 9/30/2007 22:34'!
valueString
	value isNil ifTrue: [^''].
	value isString ifTrue: [^value].
	^value printString! !

!SpreadsheetCell methodsFor: 'accessing' stamp: 'hmm 9/30/2007 20:52'!
valueString: aString
	"first simple version. Later on this should be handled by the cell's format"
	(aString notEmpty and: [aString first = $=]) ifTrue: [
		^self ruleString: (aString copyFrom: 2 to: aString size)].
	self rule: nil.
	aString isEmpty ifTrue: [
		^self value: nil].
	aString first isDigit ifTrue: [
		^self value: aString asNumber].
	self value: aString! !

!SpreadsheetCell methodsFor: 'accessing' stamp: 'hmm 9/29/2007 13:26'!
wave
	^wave! !

!SpreadsheetCell methodsFor: 'accessing' stamp: 'hmm 9/29/2007 13:26'!
wave: anInteger
	wave := anInteger! !


!SpreadsheetCell methodsFor: 'testing' stamp: 'hmm 9/29/2007 13:30'!
hasPrerequisites
	^rule notNil and: [rule cellReferences notEmpty]! !

!SpreadsheetCell methodsFor: 'testing' stamp: 'hmm 9/24/2007 21:06'!
isDummy
	"Return true if this cell should never have a value, rule or dependents"
	^rowIndex = 0! !


!SpreadsheetCell methodsFor: 'rules' stamp: 'hmm 9/29/2007 15:03'!
attachToPrerequisites
	self prerequisiteCells asSet do: [:each | each addDependent: self]! !

!SpreadsheetCell methodsFor: 'rules' stamp: 'hmm 9/29/2007 15:03'!
detachFromPrerequisites
	self prerequisiteCells do: [:each | each removeDependent: self]! !


!SpreadsheetCell methodsFor: 'updating' stamp: 'hmm 9/30/2007 21:24'!
update: aSymbol
	aSymbol = #value ifTrue: [sheet markForRecalculation: self]! !


!SpreadsheetCell methodsFor: 'recomputing' stamp: 'hmm 9/30/2007 21:29'!
recomputeValue
	self value: (rule valueInCell: self)! !


!SpreadsheetCell methodsFor: 'dependents' stamp: 'hmm 9/30/2007 21:44'!
addDependent: aDependent
	self isDummy ifFalse: [super addDependent: aDependent]! !


SequenceableCollection subclass: #SpreadsheetColumn
	instanceVariableNames: 'sheet width index'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Spreadsheet-Logic'!

!SpreadsheetColumn methodsFor: 'initialize-release' stamp: 'hmm 10/2/2007 19:30'!
index: anInteger
	index := anInteger! !

!SpreadsheetColumn methodsFor: 'initialize-release' stamp: 'hmm 9/24/2007 21:41'!
sheet: aSpreadsheet
	sheet := aSpreadsheet.
	index := 0.
	width := 70! !


!SpreadsheetColumn methodsFor: 'accessing' stamp: 'hmm 9/24/2007 21:18'!
at: anInteger
	^(sheet rows at: anInteger) at: index! !

!SpreadsheetColumn methodsFor: 'accessing' stamp: 'hmm 9/24/2007 21:18'!
at: anInteger put: anObject
	^(sheet rows at: anInteger) at: index put: anObject! !

!SpreadsheetColumn methodsFor: 'accessing' stamp: 'hmm 9/24/2007 20:49'!
size
	^sheet rows size! !

!SpreadsheetColumn methodsFor: 'accessing' stamp: 'hmm 9/24/2007 20:52'!
width
	^width! !

!SpreadsheetColumn methodsFor: 'accessing' stamp: 'hmm 10/3/2007 23:29'!
width: anInteger
	width := anInteger.
	sheet changedColumn: self! !


!SpreadsheetColumn methodsFor: 'testing' stamp: 'hmm 10/2/2007 19:31'!
isDummy
	^index = 0! !


!SpreadsheetColumn methodsFor: 'copying' stamp: 'hmm 10/2/2007 19:32'!
shallowCopy
	^self clone! !


SequenceableCollection subclass: #SpreadsheetRow
	instanceVariableNames: 'sheet cells height index'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Spreadsheet-Logic'!

!SpreadsheetRow methodsFor: 'initialize-release' stamp: 'hmm 9/29/2007 13:16'!
sheet: aSpreadsheet
	sheet := aSpreadsheet.
	index := 0.
	height := 14.
	cells := SparseArray new: aSpreadsheet columns size withAll: (SpreadsheetCell new sheet: aSpreadsheet)! !


!SpreadsheetRow methodsFor: 'accessing' stamp: 'hmm 9/24/2007 20:51'!
at: index
	^cells at: index! !

!SpreadsheetRow methodsFor: 'accessing' stamp: 'hmm 9/24/2007 20:51'!
at: index put: anObject
	^cells at: index put: anObject! !

!SpreadsheetRow methodsFor: 'accessing' stamp: 'hmm 9/30/2007 21:11'!
cells
	^cells! !

!SpreadsheetRow methodsFor: 'accessing' stamp: 'hmm 9/24/2007 20:50'!
height
	^height! !

!SpreadsheetRow methodsFor: 'accessing' stamp: 'hmm 11/10/2007 12:33'!
height: anInteger
	height := anInteger.
	sheet changedRow: self! !

!SpreadsheetRow methodsFor: 'accessing' stamp: 'hmm 9/24/2007 21:18'!
index
	^index! !

!SpreadsheetRow methodsFor: 'accessing' stamp: 'hmm 9/24/2007 21:18'!
index: anInteger
	index := anInteger! !

!SpreadsheetRow methodsFor: 'accessing' stamp: 'hmm 9/24/2007 20:51'!
size
	^cells size! !


!SpreadsheetRow methodsFor: 'copying' stamp: 'hmm 9/24/2007 21:10'!
postCopy
	cells := cells copy! !

!SpreadsheetRow methodsFor: 'copying' stamp: 'hmm 9/24/2007 21:10'!
shallowCopy
	^self clone! !


!SpreadsheetRow methodsFor: 'testing' stamp: 'hmm 9/24/2007 21:17'!
isDummy
	^index = 0! !


Object subclass: #SpreadsheetRule
	instanceVariableNames: 'expression'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Spreadsheet-Logic'!


Object subclass: #SpreadsheetRuleParser
	instanceVariableNames: 'source token tokenType result tokenBuffer cell'
	classVariableNames: 'TokenTypes'
	poolDictionaries: ''
	category: 'Spreadsheet-Logic'!

!SpreadsheetRuleParser methodsFor: 'scanning' stamp: 'hmm 9/26/2007 15:15'!
getToken
	source skipSeparators.
	source atEnd ifTrue: [
		token := nil.
		^tokenType := #eof].
	token := source next.
	tokenType := TokenTypes at: token asInteger + 1.
	tokenType first = $x ifTrue: [self perform: tokenType].
	^tokenType! !

!SpreadsheetRuleParser methodsFor: 'scanning' stamp: 'hmm 11/12/2007 23:06'!
xDigit
	tokenBuffer reset; nextPut: token.
	[source atEnd not and: [source peek isDigit]] whileTrue: [
		tokenBuffer nextPut: source next].
	(source peekFor: $.) ifTrue: [
		tokenBuffer nextPut: $..
		[source atEnd not and: [source peek isDigit]] whileTrue: [
			tokenBuffer nextPut: source next]].
	token := tokenBuffer contents.
	tokenType := #number! !

!SpreadsheetRuleParser methodsFor: 'scanning' stamp: 'hmm 10/3/2007 20:53'!
xLetter
	tokenBuffer reset; nextPut: token.
	[source atEnd not and: [source peek isDigit or: [(TokenTypes at: source peek asInteger+1) = #xLetter]]] whileTrue: [
		tokenBuffer nextPut: source next].
	token := tokenBuffer contents.
	(SpreadsheetRuleFunction isValidFunctionName: token) ifTrue: [
		^tokenType := #functionName].
	(SpreadsheetRuleCellReference isValidReference: token) ifTrue: [
		^tokenType := #cellReference].
	tokenType := #identifier! !


!SpreadsheetRuleParser methodsFor: 'parsing' stamp: 'hmm 9/30/2007 21:03'!
cellReference
	tokenType = #cellReference ifFalse: [^false].
	result := SpreadsheetRuleCellReference fromString: token.
	result makeRelativeTo: cell.
	self getToken.
	^true! !

!SpreadsheetRuleParser methodsFor: 'parsing' stamp: 'hmm 9/27/2007 08:57'!
expression
	^self sum! !

!SpreadsheetRuleParser methodsFor: 'parsing' stamp: 'hmm 9/30/2007 20:44'!
factor
	self primary ifTrue: [^true].
	tokenType = #leftParen ifTrue: [
		self getToken.
		self expression ifFalse: [self error: 'expression expected'].
		tokenType = #rightParen ifFalse: [self error: 'right parenthesis expected'].
		self getToken.
		^true].
	^false! !

!SpreadsheetRuleParser methodsFor: 'parsing' stamp: 'hmm 9/30/2007 14:50'!
function
	| func args |
	tokenType = #functionName ifFalse: [^false].
	func := token.
	self getToken.
	tokenType = #leftParenthesis ifFalse: [
		"function without args can be written without parentheses"
		result := SpreadsheetRuleFunction withName: func.
		^true].
	args := OrderedCollection new.
	self getToken.
	tokenType = #rightParenthesis ifTrue: [
		"empty open/close parentheses"
		result := SpreadsheetRuleFunction withName: func.
		^true].
	[self expression ifFalse: [self expected: #expression].
	args add: result.
	tokenType = #comma] whileTrue: [self getToken].
	tokenType = #rightParenthesis ifFalse: [self expected: 'comma or right parenthesis'].
	self getToken.
	result := SpreadsheetRuleFunction withName: func.
	result arguments: args.
	^true! !

!SpreadsheetRuleParser methodsFor: 'parsing' stamp: 'hmm 9/30/2007 14:44'!
number
	tokenType = #number ifTrue: [
		result := SpreadsheetRuleConstant value: token asNumber.
		self getToken.
		^true].
	^false! !

!SpreadsheetRuleParser methodsFor: 'parsing' stamp: 'hmm 9/30/2007 21:09'!
parse: aString
	^self parse: aString forCell: (SpreadsheetCell new point: 1@1)! !

!SpreadsheetRuleParser methodsFor: 'parsing' stamp: 'hmm 9/30/2007 21:09'!
parse: aString forCell: aCell
	cell := aCell.
	source := aString readStream.
	tokenBuffer := WriteStream on: (String new: 20).
	self getToken.
	self expression ifFalse: [self error: 'expression expected'].
	tokenType = #eof ifFalse: [self error: 'end of input expected'].
	^result! !

!SpreadsheetRuleParser methodsFor: 'parsing' stamp: 'hmm 9/30/2007 20:47'!
primary
	self number ifTrue: [^true].
	self cellReference ifTrue: [^true].
	self function ifTrue: [^true].
	^false! !

!SpreadsheetRuleParser methodsFor: 'parsing' stamp: 'hmm 9/30/2007 20:03'!
sum
	| args operator function |
	self term ifFalse: [^false].
	tokenType = #addOp ifFalse: [^true].
	args := OrderedCollection with: result.
	operator := token.
	[tokenType = #addOp] whileTrue: [
		token = operator ifFalse: [
			function := SpreadsheetRuleFunction withName: operator asString.
			function arguments: args.
			args := OrderedCollection with: function.
			operator := token].
		self getToken.
		self term ifFalse: [self error: 'term expected'].
		args add: result].
	args size = 1
		ifTrue: [result := args first]
		ifFalse: [
			result := SpreadsheetRuleFunction withName: operator asString.
			result arguments: args].
	^true! !

!SpreadsheetRuleParser methodsFor: 'parsing' stamp: 'hmm 9/30/2007 20:04'!
term
	| args operator function |
	self factor ifFalse: [^false].
	tokenType = #mulOp ifFalse: [^true].
	args := OrderedCollection with: result.
	operator := token.
	[tokenType = #mulOp] whileTrue: [
		token = operator ifFalse: [
			function := SpreadsheetRuleFunction withName: operator asString.
			function arguments: args.
			args := OrderedCollection with: function.
			operator := token].
		self getToken.
		self factor ifFalse: [self error: 'factor expected'].
		args add: result].
	args size = 1
		ifTrue: [result := args first]
		ifFalse: [
			result := SpreadsheetRuleFunction withName: operator asString.
			result arguments: args].
	^true! !

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

SpreadsheetRuleParser class
	instanceVariableNames: ''!

!SpreadsheetRuleParser class methodsFor: 'class initialization' stamp: 'hmm 9/30/2007 20:10'!
initialize
	"self initialize"
	TokenTypes := Array new: 256 withAll: #undefined.
	self tokenType: #xLetter for: $A to: $Z.
	self tokenType: #xLetter for: $a to: $z.
	self tokenType: #xLetter for: $$.
	self tokenType: #xDigit for: $0 to: $9.
	self tokenType: #xQuote forAll: '''"'.
	self tokenType: #leftParen for: $(.
	self tokenType: #rightParen for: $).
	self tokenType: #addOp forAll: '+-'.
	self tokenType: #mulOp forAll: '*/'.
	! !

!SpreadsheetRuleParser class methodsFor: 'class initialization' stamp: 'hmm 9/26/2007 15:27'!
tokenType: aSymbol for: aCharacter
	TokenTypes at: aCharacter asInteger+1 put: aSymbol! !

!SpreadsheetRuleParser class methodsFor: 'class initialization' stamp: 'hmm 9/26/2007 15:28'!
tokenType: aSymbol for: firstCharacter to: lastCharacter
	firstCharacter asInteger to: lastCharacter asInteger do: [:i | TokenTypes at:i+1 put: aSymbol]! !

!SpreadsheetRuleParser class methodsFor: 'class initialization' stamp: 'hmm 9/26/2007 15:28'!
tokenType: aSymbol forAll: aString
	aString do: [:char | TokenTypes at: char asInteger+1 put: aSymbol]! !


Object subclass: #SpreadsheetRulePart
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Spreadsheet-Logic'!

!SpreadsheetRulePart methodsFor: 'evaluating' stamp: 'hmm 9/29/2007 14:54'!
valueInCell: aCell
	self subclassResponsibility! !


!SpreadsheetRulePart methodsFor: 'accessing' stamp: 'hmm 9/30/2007 21:15'!
cellReferences
	"subclasses with direct or indirect cell references should return these"
	^#()
	! !

!SpreadsheetRulePart methodsFor: 'accessing' stamp: 'hmm 9/29/2007 15:05'!
prerequisiteCellsFor: aCell
	"subclasses with direct or indirect cell references should return these"
	^#()
	! !


!SpreadsheetRulePart methodsFor: 'printing' stamp: 'hmm 10/3/2007 20:55'!
printOn: aStream
	aStream nextPut: ${.
	self printOn: aStream for: nil precedence: 0.
	aStream nextPut: $}! !

!SpreadsheetRulePart methodsFor: 'printing' stamp: 'hmm 10/2/2007 19:35'!
printOn: aStream for: aCell
	aStream nextPut: ${.
	self printOn: aStream for: aCell precedence: 0.
	aStream nextPut: $}! !

!SpreadsheetRulePart methodsFor: 'printing' stamp: 'hmm 10/2/2007 19:36'!
printOn: aStream for: aCell precedence: precedence
	self subclassResponsibility! !


SpreadsheetRulePart subclass: #SpreadsheetRuleCellReference
	instanceVariableNames: 'rowIndex columnIndex rowBase columnBase'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Spreadsheet-Logic'!

!SpreadsheetRuleCellReference methodsFor: 'evaluating' stamp: 'hmm 11/3/2007 11:28'!
referenceInCell: aCell
	| refPoint |
	refPoint := (columnBase isNil ifTrue: [columnIndex] ifFalse: [aCell columnIndex - columnBase + columnIndex])
				@ (rowBase isNil ifTrue: [rowIndex] ifFalse: [aCell rowIndex - rowBase + rowIndex]).
	^aCell sheet realCellAt: refPoint! !

!SpreadsheetRuleCellReference methodsFor: 'evaluating' stamp: 'hmm 9/29/2007 15:07'!
valueInCell: aCell
	^(self referenceInCell: aCell) value
	! !


!SpreadsheetRuleCellReference methodsFor: 'accessing' stamp: 'hmm 9/30/2007 21:14'!
cellReferences
	^Array with: self! !

!SpreadsheetRuleCellReference methodsFor: 'accessing' stamp: 'hmm 9/30/2007 15:06'!
precedence
	^5! !

!SpreadsheetRuleCellReference methodsFor: 'accessing' stamp: 'hmm 9/29/2007 15:06'!
prerequisiteCellsFor: aCell
	^Array with: (self referenceInCell: aCell)! !



!SpreadsheetRuleCellReference methodsFor: 'initialize-release' stamp: 'hmm 9/30/2007 21:04'!
makeRelativeTo: aCell
	rowBase isNil ifFalse: [rowBase := aCell rowIndex].
	columnBase isNil ifFalse: [columnBase := aCell columnIndex].! !

!SpreadsheetRuleCellReference methodsFor: 'initialize-release' stamp: 'hmm 9/30/2007 21:05'!
row: row relative: rowRelativeBool column: column relative: columnRelativeBool
	rowIndex := row.
	rowBase := rowRelativeBool ifTrue: [1].
	columnIndex := column.
	columnBase := columnRelativeBool ifTrue: [1]! !


!SpreadsheetRuleCellReference methodsFor: 'printing' stamp: 'hmm 10/2/2007 19:39'!
printOn: aStream for: aCell precedence: ignored
	| t c r |
	c := columnIndex.
	r := rowIndex.
	columnBase isNil
		ifTrue: [aStream nextPut: $$]
		ifFalse: [aCell notNil ifTrue: [c := c - columnBase + aCell columnIndex]].
	t := c -1 // 26.
	t > 0 ifTrue: [aStream nextPut: (Character value: $A asInteger -1 + t)].
	aStream nextPut: (Character value: $A asInteger + (c -1 \\ 26)).
	rowBase isNil
		ifTrue: [aStream nextPut: $$]
		ifFalse:  [aCell notNil ifTrue: [r := r - rowBase + aCell rowIndex]]..
	aStream print: r.! !

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

SpreadsheetRuleCellReference class
	instanceVariableNames: ''!

!SpreadsheetRuleCellReference class methodsFor: 'instance creation' stamp: 'hmm 9/30/2007 20:47'!
fromString: aString
	| str columnAbsolute column rowAbsolute row |
	str := aString asUppercase readStream.
	columnAbsolute := str peekFor: $$.
	column := 0.
	(str peek notNil and: [str peek isLetter]) ifFalse: [^nil].
	[str atEnd not and: [str peek >= $A and: [str peek <= $Z]]] whileTrue: [
		column := column * 26 + (str next asInteger - $A asInteger)].
	column := column+1.
	(column between: 1 and: 256) ifFalse: [^nil].
	rowAbsolute := str peekFor: $$.
	row := 0.
	[str atEnd not and: [str peek >= $0 and: [str peek <= $9]]] whileTrue: [
		row := row * 10 + (str next asInteger - $0 asInteger)].
	(row between: 1 and: 65536) ifFalse: [^nil].
	^self new row: row relative: rowAbsolute not column: column relative: columnAbsolute not! !

!SpreadsheetRuleCellReference class methodsFor: 'instance creation' stamp: 'hmm 9/30/2007 20:41'!
isValidReference: aString
	^(self fromString: aString) notNil! !


SpreadsheetRulePart subclass: #SpreadsheetRuleConstant
	instanceVariableNames: 'value'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Spreadsheet-Logic'!

!SpreadsheetRuleConstant methodsFor: 'evaluating' stamp: 'hmm 9/29/2007 14:54'!
valueInCell: aCell
	^value! !


!SpreadsheetRuleConstant methodsFor: 'accessing' stamp: 'hmm 9/30/2007 15:06'!
precedence
	^5! !

!SpreadsheetRuleConstant methodsFor: 'accessing' stamp: 'hmm 9/30/2007 15:02'!
value: anObject
	value := anObject! !


!SpreadsheetRuleConstant methodsFor: 'printing' stamp: 'hmm 10/2/2007 19:36'!
printOn: aStream for: aCell precedence: precedence
	value printOn: aStream! !

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

SpreadsheetRuleConstant class
	instanceVariableNames: ''!

!SpreadsheetRuleConstant class methodsFor: 'instance creation' stamp: 'hmm 9/30/2007 15:03'!
value: anObject
	^self new value: anObject! !


SpreadsheetRulePart subclass: #SpreadsheetRuleFunction
	instanceVariableNames: 'selector arguments'
	classVariableNames: 'FunctionNames Precedences'
	poolDictionaries: ''
	category: 'Spreadsheet-Logic'!

!SpreadsheetRuleFunction methodsFor: 'evaluating' stamp: 'hmm 11/3/2007 11:30'!
valueInCell: aCell
	^[self perform: selector with: (arguments collect: [:each | each valueInCell: aCell])]
		on: Error
		do: [:ex | ex return: #Error]! !


!SpreadsheetRuleFunction methodsFor: 'functions' stamp: 'hmm 10/3/2007 20:46'!
date: aCell
	^Date today! !

!SpreadsheetRuleFunction methodsFor: 'functions' stamp: 'hmm 9/30/2007 20:15'!
div: operandValues
	| result |
	result := operandValues first.
	2 to: operandValues size do: [:i | result := result / (operandValues at: i)].
	^result! !

!SpreadsheetRuleFunction methodsFor: 'functions' stamp: 'hmm 9/30/2007 20:15'!
minus: operandValues
	| result |
	result := operandValues first.
	2 to: operandValues size do: [:i | result := result - (operandValues at: i)].
	^result! !

!SpreadsheetRuleFunction methodsFor: 'functions' stamp: 'hmm 9/30/2007 20:15'!
mul: operandValues
	| result |
	result := operandValues first.
	2 to: operandValues size do: [:i | result := result * (operandValues at: i)].
	^result! !

!SpreadsheetRuleFunction methodsFor: 'functions' stamp: 'hmm 9/30/2007 20:15'!
plus: operandValues
	| result |
	result := operandValues first.
	2 to: operandValues size do: [:i | result := result + (operandValues at: i)].
	^result! !


!SpreadsheetRuleFunction methodsFor: 'accessing' stamp: 'hmm 9/30/2007 20:02'!
arguments: aCollection
	arguments := aCollection! !

!SpreadsheetRuleFunction methodsFor: 'accessing' stamp: 'hmm 9/30/2007 21:15'!
cellReferences
	| result |
	result := OrderedCollection new.
	arguments do: [:each | result addAll: each cellReferences].
	^result! !

!SpreadsheetRuleFunction methodsFor: 'accessing' stamp: 'hmm 9/30/2007 15:10'!
initialize
	arguments := #()! !

!SpreadsheetRuleFunction methodsFor: 'accessing' stamp: 'hmm 9/30/2007 20:13'!
precedence
	^Precedences at: (FunctionNames at: selector) ifAbsent: [5]! !

!SpreadsheetRuleFunction methodsFor: 'accessing' stamp: 'hmm 9/30/2007 14:51'!
prerequisiteCellsFor: aCell
	| result |
	result := aCell.
	arguments do: [:each | result addAll: (each prerequisiteCellsFor: aCell)].
	^result! !

!SpreadsheetRuleFunction methodsFor: 'accessing' stamp: 'hmm 9/30/2007 15:10'!
selector: aSymbol
	selector := aSymbol! !


!SpreadsheetRuleFunction methodsFor: 'printing' stamp: 'hmm 10/3/2007 20:55'!
printOn: aStream for: aCell precedence: outerPrecedence
	| name |
	name := FunctionNames at: selector.
	name first isLetter
		ifTrue: [
			aStream nextPutAll: name; nextPut: $(.
			arguments
				do: [:each | each printOn: aStream  for: aCell precedence: 0]
				separatedBy: [aStream nextPutAll: ', '].
			aStream nextPut: $)]
		ifFalse: [
			outerPrecedence > self precedence ifTrue: [aStream nextPut: $(].
			arguments first printOn: aStream for: aCell precedence: self precedence.
			2 to: arguments size do: [:i |
				aStream nextPutAll: name.
				(arguments at: i) printOn: aStream for: aCell precedence: self precedence+1].
			outerPrecedence > self precedence ifTrue: [aStream nextPut: $)]]! !

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

SpreadsheetRuleFunction class
	instanceVariableNames: ''!

!SpreadsheetRuleFunction class methodsFor: 'class initialization' stamp: 'hmm 10/3/2007 20:48'!
initialize
	"SpreadsheetRuleFunction initialize"
	FunctionNames := Dictionary new.
	Precedences := Dictionary new.
	self initializeFunctions.
	self initializeOperators! !

!SpreadsheetRuleFunction class methodsFor: 'class initialization' stamp: 'hmm 10/3/2007 20:48'!
initializeFunctions
	"SpreadsheetRuleFunction initializeFunctions"
	| selector |
	#('DATE')
	do: [:string |
		selector := (string asLowercase, ':') asSymbol.
		FunctionNames at: string put: selector.
		FunctionNames at: selector put: string]! !

!SpreadsheetRuleFunction class methodsFor: 'class initialization' stamp: 'hmm 10/3/2007 20:46'!
initializeOperators
	"SpreadsheetRuleFunction initializeOperators"
	#(
		('+' #plus:)
		('-' #minus:)
		('*' #mul:)
		('/' #div:))
	do: [:pair |
		FunctionNames at: pair first put: pair last.
		FunctionNames at: pair last put: pair first].
	'+-' do: [:each |
		Precedences at: (String with: each) put: 3].
	'*/' do: [:each |
		Precedences at: (String with: each) put: 4].
	! !


!SpreadsheetRuleFunction class methodsFor: 'instance creation' stamp: 'hmm 10/3/2007 20:53'!
withName: aString
	^self new selector: (FunctionNames at: aString asUppercase)! !


!SpreadsheetRuleFunction class methodsFor: 'testing' stamp: 'hmm 9/30/2007 20:22'!
isValidFunctionName: aString
	^FunctionNames includesKey: aString asUppercase! !

SpreadsheetRuleParser initialize!
SpreadsheetRuleFunction initialize!