TextMorph subclass: #SpreadsheetCellTextMorph
	instanceVariableNames: 'cell commitBlock'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Spreadsheet-Morphic'!


!SpreadsheetCellTextMorph methodsFor: 'geometry' stamp: 'hmm 11/2/2007 22:59'!
minimumExtent
	| minExt |
	textStyle ifNil: [^ 9@16].
	borderWidth ifNil: [^ 9@16].
	minExt _ (9@textStyle lineGrid) + (borderWidth*2).
	margins ifNil: [^ minExt].
	^ ((0@0 extent: minExt) expandBy: margins) extent! !


!SpreadsheetCellTextMorph methodsFor: 'initialization' stamp: 'hmm 11/10/2007 13:55'!
initialize
	super initialize.
	self autoFit: false.
	self borderWidth: 2; borderColor: Color black.
	self backgroundColor: Color white.
	textStyle := textStyle copy
				firstIndent: 2;
				restIndent: 2;
				lineGrid: 13;
				leading: 1! !


!SpreadsheetCellTextMorph methodsFor: 'drawing' stamp: 'hmm 11/2/2007 23:32'!
drawNullTextOn: aCanvas
	aCanvas fillRectangle: self bounds fillStyle: self fillStyle borderStyle: self borderStyle.
! !


!SpreadsheetCellTextMorph methodsFor: 'private' stamp: 'hmm 11/2/2007 23:34'!
fit
	"Adjust my bounds to fit the text.  Should be a no-op if autoFit is not specified.
	Required after the text changes,
	or if wrapFlag is true and the user attempts to change the extent."

	| newExtent para cBounds lastOfLines heightOfLast |
	self isAutoFit 
		ifTrue: 
			[newExtent := (self paragraph extent max: 9 @ textStyle lineGrid) + (0 @ 1).
			newExtent := newExtent + (2 * borderWidth).
			margins 
				ifNotNil: [newExtent := ((0 @ 0 extent: newExtent) expandBy: margins) extent].
			newExtent ~= bounds extent 
				ifTrue: 
					[(container isNil and: [successor isNil]) 
						ifTrue: 
							[para := paragraph.	"Save para (layoutChanged smashes it)"
							super extent: newExtent.
							paragraph := para]].
			container notNil & successor isNil 
				ifTrue: 
					[cBounds := container bounds truncated.
					"23 sept 2000 - try to allow vertical growth"
					lastOfLines := self paragraph lines last.
					heightOfLast := lastOfLines bottom - lastOfLines top.
					(lastOfLines last < text size 
						and: [lastOfLines bottom + heightOfLast >= self bottom]) 
							ifTrue: 
								[container releaseCachedState.
								cBounds := cBounds origin corner: cBounds corner + (0 @ heightOfLast)].
					self privateBounds: cBounds]].

	"These statements should be pushed back into senders"
	self paragraph positionWhenComposed: self position.
	successor ifNotNil: [successor predecessorChanged].
	self changed	"Too conservative: only paragraph composition
					should cause invalidation."! !


!SpreadsheetCellTextMorph methodsFor: 'editing' stamp: 'hmm 11/3/2007 11:16'!
acceptContents
	super acceptContents.
	owner acceptEditedCellContents: text asString! !

!SpreadsheetCellTextMorph methodsFor: 'editing' stamp: 'hmm 11/2/2007 23:40'!
acceptOnCR
	^true! !

!SpreadsheetCellTextMorph methodsFor: 'editing' stamp: 'hmm 11/30/2007 20:28'!
keyStroke: evt
	| char |
	char := evt keyCharacter.
	char = Character cr ifTrue: [
		self acceptContents.
		^owner moveEditCellNextLine].
	char = Character tab ifTrue: [
		self acceptContents.
		^owner moveEditCellTab: evt shiftPressed not].
	^super keyStroke: evt! !


!SpreadsheetCellTextMorph methodsFor: 'event handling' stamp: 'hmm 11/12/2007 22:22'!
keyboardFocusChange: aBoolean
	aBoolean ifTrue: [^super keyboardFocusChange: aBoolean].
	owner deleteEditor! !


Morph subclass: #SpreadsheetColumnLabelsMorph
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Spreadsheet-Morphic'!


Morph subclass: #SpreadsheetGridMorph
	instanceVariableNames: 'spreadsheet firstVisible editor editedCell'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Spreadsheet-Morphic'!

!SpreadsheetGridMorph methodsFor: 'drawing' stamp: 'hmm 11/2/2007 23:00'!
drawOn: aCanvas
	| r rowRange columnRange cellRect |
	aCanvas clipBy: self bounds during: [:canvas |
		r := canvas clipRect translateBy: self position negated.
		rowRange := spreadsheet visibleRowsIn: r.
		columnRange := spreadsheet visibleColumnsIn: r.
		cellRect := self rectangleForCellRange: columnRange first @ rowRange first to: columnRange last @ rowRange last.
		"Transcript cr; print: r.
		Transcript cr; tab; print: rect.
		Transcript cr; tab; print: rowRange; space; print: columnRange.
		Transcript cr; tab; print: canvas origin; endEntry."
		self
			drawOn: canvas at: cellRect origin
			rowRange: rowRange
			columnRange: columnRange
			cellsDo: [:cell :rect |
				canvas clipBy: rect during: [:sc | sc drawString: cell valueString at: rect origin + (2@1)]]]! !

!SpreadsheetGridMorph methodsFor: 'drawing' stamp: 'hmm 10/3/2007 19:45'!
drawOn: aCanvas at: aPoint rowRange: rowRange columnRange: columnRange cellsDo: aBlock
	"draw all cells and their contents with borders"
	| h w borderColor y x row column cell |
	h := 1.
	rowRange do: [:i | h := h + (spreadsheet rows at: i) height].
	w := 1.
	columnRange do: [:i | w := w + (spreadsheet columns at: i) width].
	aCanvas fillRectangle: (aPoint extent: w@h) color: Color white.
	borderColor := Color gray.
	y := aPoint y.
	rowRange do: [:r |
		aCanvas fillRectangle: (aPoint x @ y extent: w@1) color: borderColor.
		y := y + (spreadsheet rows at: r) height].
	aCanvas fillRectangle: (aPoint x @ y extent: w@1) color: borderColor.
	x := aPoint x.
	columnRange do: [:c |
		aCanvas fillRectangle: (x @ aPoint y extent: 1@h) color: borderColor.
		x := x + (spreadsheet columns at: c) width].
	aCanvas fillRectangle: (x @ aPoint y extent: 1@h) color: borderColor.
	y := aPoint y + 1.
	rowRange do: [:r |
		row := spreadsheet rows at: r.
		x := aPoint x + 1.
		columnRange do: [:c |
			column := spreadsheet columns at: c.
			cell := row at: c.
			cell isDummy ifFalse: [
				aBlock value: cell value: (x@y extent: (column width-1 @ (row height-1)))].
			x := x + column width].
		y := y + row height]! !


!SpreadsheetGridMorph methodsFor: 'initialization' stamp: 'hmm 9/30/2007 21:36'!
initialize
	super initialize.
	spreadsheet := Spreadsheet new.
	spreadsheet addDependent: self! !



!SpreadsheetGridMorph methodsFor: 'updating' stamp: 'hmm 11/10/2007 12:37'!
update: aSymbol
	aSymbol = #displayFull ifTrue: [
		self repositionEditor.
		^self changed]! !

!SpreadsheetGridMorph methodsFor: 'updating' stamp: 'hmm 11/10/2007 12:37'!
update: aSymbol with: anObject
	aSymbol = #display ifTrue: [
		self repositionEditor.
		^self invalidRect: (self rectangleForCellRange: anObject origin to: anObject corner)]! !


!SpreadsheetGridMorph methodsFor: 'private' stamp: 'hmm 11/3/2007 11:22'!
deleteEditor
	editor isNil ifFalse: [
		editor delete.
		editor := editedCell := nil]! !

!SpreadsheetGridMorph methodsFor: 'private' stamp: 'hmm 10/3/2007 19:51'!
rectangleForCellRange: originPoint to: cornerPoint
	^((spreadsheet originForCellPoint: originPoint) corner: (spreadsheet cornerForCellPoint: cornerPoint)) translateBy: self position! !

!SpreadsheetGridMorph methodsFor: 'private' stamp: 'hmm 11/10/2007 12:38'!
repositionEditor
	editor isNil ifFalse: [
		editor bounds: ((self rectangleForCellRange: editedCell to: editedCell) expandBy: (1@1 corner: 2@2))]! !


!SpreadsheetGridMorph methodsFor: 'event handling' stamp: 'hmm 11/3/2007 11:16'!
acceptEditedCellContents: aString
	spreadsheet cellStringAt: editedCell put: aString! !

!SpreadsheetGridMorph methodsFor: 'event handling' stamp: 'hmm 11/10/2007 13:58'!
editCell: aPoint evt: anEvent
	editor isNil ifFalse: [self acceptEditedCellContents: editor contents asString].
	self deleteEditor.
	editedCell := aPoint.
	editor := SpreadsheetCellTextMorph new.
	editor borderWidth: 2.
	self repositionEditor.
	editor contentsWrapped: (spreadsheet cellRuleOrValueStringAt: aPoint).
	"editor autoFit: true."
	self addMorph: editor.
	anEvent isNil
		ifTrue: [ActiveHand newKeyboardFocus: editor]
		ifFalse: [
			ActiveHand newMouseFocus: editor.
			editor mouseDown: anEvent]! !

!SpreadsheetGridMorph methodsFor: 'event handling' stamp: 'hmm 11/2/2007 23:04'!
handlesMouseDown: evt
	^evt redButtonPressed! !

!SpreadsheetGridMorph methodsFor: 'event handling' stamp: 'hmm 11/3/2007 11:15'!
mouseDown: evt
	| relative rect rowRange columnRange |
	evt redButtonPressed ifTrue: [
		relative := evt position - self position.
		rect := relative extent: 1@1.
		rowRange := spreadsheet visibleRowsIn: rect.
		columnRange := spreadsheet visibleColumnsIn: rect.
		self editCell: columnRange first @ rowRange first evt: evt]! !

!SpreadsheetGridMorph methodsFor: 'event handling' stamp: 'hmm 11/3/2007 11:17'!
moveEditCellNextLine
	self editCell: editedCell + (0@1) evt: nil! !


!SpreadsheetGridMorph methodsFor: 'opening' stamp: 'hmm 12/2/2007 17:54'!
openInWindow
	"SpreadsheetGridMorph new openInWindow"
	| window scrollPane |
	window := SystemWindow labelled: 'Spreadsheet'.
	scrollPane := ScrollPane new.
	scrollPane scroller offset: 0@0; addMorph: (self extent: 500@500).
	scrollPane alwaysShowHScrollBar: true.
	window addMorph: scrollPane fullFrame: ((LayoutFrame fractions: (0@0 extent: 1@1)) topOffset: 30).
	window openInWorldExtent: 400@300! !


Morph subclass: #SpreadsheetMorph
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Spreadsheet-Morphic'!

!SpreadsheetMorph methodsFor: 'opening' stamp: 'hmm 12/2/2007 17:49'!
openInWindow
	"SpreadsheetMorph new openInWindow"
	| window |
	window := SystemWindow labelled: 'Spreadsheet'.
	window addMorph: self fullFrame: ((LayoutFrame fractions: (0@0 extent: 1@1)) topOffset: 30).
	window openInWorldExtent: 400@300! !
